/* Instruction printing code for the OpenRISC 1000
   Copyright (C) 1990, 93, 94, 95, 1998 Free Software Foundation, Inc.
   Contributed by Damjan Lampret <lampret@opencores.org>.
   Modified from a29k port.

This file is part of GDB.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#define NO_RELOC 20
#define RELOC_32 1
#define RELOC_8 2
#define RELOC_CONST 3
#define RELOC_CONSTH 4

#define DEBUG 0

#define EXTEND12(x) ((x) & 0x1000 ? ((x) | 0xffffe000) : ((x) & ~0x1000))
#define EXTEND16(x) ((x) & 0x8000 ? ((x) | 0xffff0000) : ((x) & ~0x8000))
#define EXTEND26(x) ((x) & 0x8000000 ? ((x) | 0xf0000000) : ((x)))

#include "dis-asm.h"
#include "opcode/or16.h"

#include <ctype.h>
#include <string.h>

/* Now find the four bytes of INSN_CH and put them in *INSN.  */
static void
find_bytes_big (insn_ch, insn)
     unsigned char *insn_ch;
     unsigned long *insn;
{
  *insn = ((unsigned long)insn_ch[0] << 24) +
          ((unsigned long)insn_ch[1] << 16) +
          ((unsigned long)insn_ch[2] << 8) +
          (unsigned long)insn_ch[3];
#if DEBUG
  printf ("find_bytes_big3: %x\n", *insn);
#endif
}

static void
find_bytes_little (insn_ch, insn)
     unsigned char *insn_ch;
     unsigned long *insn;
{
  *insn = ((unsigned long)insn_ch[3] << 24) +
          ((unsigned long)insn_ch[2] << 16) +
          ((unsigned long)insn_ch[1] << 8) +
          (unsigned long)insn_ch[0];
}


typedef void (*find_byte_func_type)
     PARAMS ((unsigned char *, unsigned long *));

static unsigned long
or16_extract(param_ch, enc_initial, insn)
     char param_ch;
     char *enc_initial;
     unsigned long insn;
{
  char *enc;
  unsigned long ret = 0;
  int opc_pos = 0;
  int param_pos = 0;

  for (enc = enc_initial; *enc != '\0'; enc++)
    if (*enc == param_ch)
      if (enc - 2 >= enc_initial && (*(enc - 2) == '0') && (*(enc - 1) == 'x'))
        continue;
      else
        param_pos++;

#if DEBUG
  printf("or16_extract: %x ", param_pos);
#endif
  opc_pos = 32;
  for (enc = enc_initial; *enc != '\0'; )
    if ((*enc == '0') && (*(enc+1) == 'x')) 
      {
        opc_pos -= 4;
        if ((param_ch == '0') || (param_ch == '1')) 
          {
            unsigned long tmp = strtol(enc, NULL, 16);
#if DEBUG
            printf(" enc=%s, tmp=%x ", enc, tmp);
#endif
            if (param_ch == '0')
              tmp = 15 - tmp;
            ret |= tmp << opc_pos;
          }
        enc += 3;
      }
    else if ((*enc == '0') || (*enc == '1')) 
      {
        opc_pos--;
        if (param_ch == *enc)
          ret |= 1 << opc_pos;
        enc++;
      }
    else if (*enc == param_ch) 
      {
        opc_pos--;
        param_pos--;
#if DEBUG
        printf("\n  ret=%x opc_pos=%x, param_pos=%x\n", ret, opc_pos, param_pos);
#endif  
        if (islower(param_ch))
          ret -= ((insn >> opc_pos) & 0x1) << param_pos;
        else
          ret += ((insn >> opc_pos) & 0x1) << param_pos;
        enc++;
      }
    else if (isalpha(*enc)) 
      {
        opc_pos--;
        enc++;
      }
    else if (*enc == '-') 
      {
        opc_pos--;
        enc++;
      }
    else
      enc++;

#if DEBUG
  printf ("ret=%x\n", ret);
#endif
  return ret;
}

static int
or16_opcode_match (insn, encoding)
     unsigned long insn;
     char *encoding;
{
  unsigned long ones, zeros;

#if DEBUG
  printf ("or16_opcode_match: %.8lx\n", insn);
#endif    
  ones  = or16_extract('1', encoding, insn);
  zeros = or16_extract('0', encoding, insn);
  
#if DEBUG
  printf("ones: %x \n", ones);
  printf("zeros: %x \n", zeros);
#endif
  if ((insn & ones) != ones) 
    {
#if DEBUG
      printf("ret1\n");
#endif
      return 0;
    }
    
  if ((~insn & zeros) != zeros) 
    {
#if DEBUG
      printf("ret2\n");
#endif
      return 0;
    }
  
#if DEBUG
  printf ("ret3\n");
#endif
  return 1;
}


/* Print register to INFO->STREAM. Used only by print_insn. */

static void
or16_print_register (param_ch, encoding, insn, info)
     char param_ch;
     char *encoding;
     unsigned long insn;
     struct disassemble_info *info;
{
  int regnum = or16_extract(param_ch, encoding, insn);
  
#if DEBUG
  printf("or16_print_register: %c, %s, %x\n", param_ch, encoding, insn);
#endif  
#if 0
  if (param_ch == 'A')
    (*info->fprintf_func) (info->stream, "a%d", regnum);
  else if (param_ch == 'B')
    (*info->fprintf_func) (info->stream, "b%d", regnum);
  else if (param_ch == 'S')
    (*info->fprintf_func) (info->stream, "s%d", regnum);
  else if (regnum < 16)
    (*info->fprintf_func) (info->stream, "a%d", regnum);
  else if (regnum < 32)
    (*info->fprintf_func) (info->stream, "b%d", regnum-16);
  else
    (*info->fprintf_func) (info->stream, "X%d", regnum);
#else
  (*info->fprintf_func) (info->stream, "%%r%d", regnum);
#endif
}

/* Print immediate to INFO->STREAM. Used only by print_insn. */

static void
or16_print_immediate (param_ch, encoding, insn, info)
     char param_ch;
     char *encoding;
     unsigned long insn;
     struct disassemble_info *info;
{
  int imm = or16_extract (param_ch, encoding, insn);

  switch (param_ch)
    {
    case 'J':
      imm = EXTEND16(imm);

      if ((int)imm >= 0x1000)
        (*info->fprintf_func) (info->stream, "%#x", imm);
      else
        (*info->fprintf_func) (info->stream, "%d", imm);
      break;

    default:
      if (islower(param_ch))
        (*info->fprintf_func) (info->stream, "%d", imm);
      else
        (*info->fprintf_func) (info->stream, "%u", imm);
      break;
    }
}

/* Print one instruction from MEMADDR on INFO->STREAM.
   Return the size of the instruction (2 or 4 on or16).  */
static int
print_insn (memaddr, info)
     bfd_vma memaddr;
     struct disassemble_info *info;
{
  /* The raw instruction.  */
  unsigned char insn_ch[4];

  /* Address. Will be sign extened 12-bit */
  int addr;

  /* The four bytes of the instruction.  */
  unsigned long insn;

  find_byte_func_type find_byte_func = (find_byte_func_type)info->private_data;

  struct or16_opcode CONST * opcode;

  {
    int status =
      (*info->read_memory_func) (memaddr, (bfd_byte *) &insn_ch[0], 2, info);
    if (status != 0)
      {
        (*info->memory_error_func) (status, memaddr, info);
        return -1;
      }
  }

  (*find_byte_func) (&insn_ch[0], &insn);

  for (opcode = &or16_opcodes[0];
       opcode < &or16_opcodes[num_opcodes];
       ++opcode)
    {
      if (or16_opcode_match (insn, opcode->encoding))
        {
          char *s;
    
          if (insn_len(opcode->name) != 2)
            {
              int status =
                (*info->read_memory_func) (memaddr, (bfd_byte *) &insn_ch[0], 
                                           insn_len(opcode->name), info);
              if (status != 0)
                {
                  (*info->memory_error_func) (status, memaddr, info);
                  return -1;
                }

              (*find_byte_func) (&insn_ch[0], &insn);
            }

          (*info->fprintf_func) (info->stream, "%s ", opcode->name);
          for (s = opcode->args; *s != '\0'; ++s)
            {
              switch (*s)
                {
                case '\0':
                  return 4;
      
                case 'r':
                  or16_print_register(*++s, opcode->encoding, insn, info);
                  break;

                case 'X':
                  /* Extract our 12-bit address */
                  addr = or16_extract ('X', opcode->encoding, insn) << 1;

                  /* Calulate the correct address.  XXX is this really correct ?? */
                  addr = memaddr + EXTEND12(addr);

                  (*info->print_address_func)
                    (addr, info);
                  break;

                case 'Y':
                  /* Extract our 26-bit address */
                  addr = or16_extract ('Y', opcode->encoding, insn) << 1;

                  /* Calulate the correct address.  XXX is this really correct ?? */
                  addr = memaddr + EXTEND26(addr);

                  (*info->print_address_func)
                    (addr, info);
                  break;
                  
                default:
                  if (strchr (opcode->encoding, *s))
                    or16_print_immediate (*s, opcode->encoding, insn, info);
                  else
                    (*info->fprintf_func) (info->stream, "%c", *s);
                }
            }

          return insn_len(opcode->name);
        }
    }

  /* This used to be %8x for binutils.  */
  (*info->fprintf_func)
    (info->stream, ".word 0x%08x", insn);
  return 4;
}

/* Disassemble a big-endian or16 instruction.  */
int
print_insn_big_or16 (memaddr, info)
     bfd_vma memaddr;
     struct disassemble_info *info;
{
  info->private_data = (PTR) find_bytes_big;
  return print_insn (memaddr, info);
}

/* Disassemble a little-endian or16 instruction.  */
int
print_insn_little_or16 (memaddr, info)
     bfd_vma memaddr;
     struct disassemble_info *info;
{
  info->private_data = (PTR) find_bytes_little;
  return print_insn (memaddr, info);
}
